/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*******************************************************************************
 Program name:     lm_bayes
                   (Version 2.6: MORGAN compatible)
                   ----------------------------
  Author:          Andrew W. George

  Date:            5th February, 2002

  Keywords:        Pseudo-Bayesian inference, lod scores, LM-samplers,
                   meiosis indicators, sequential imputation,
                   Metropolis-Hastings restarts,
                   Rao-Blackwellization estimators.

  Purpose:         To calculate multipoint lod scores from data observed on
                   general pedigrees using a MCMC method which is less
                   susceptible to mixing problems.

  Methods
  Utilized:        LM samplers, M-H and Gibbs samplers, sequential imputation
                   for generating starting marker configurations,
                   Metropolis-Hastings restarts via sequential imputation,
                   Rao-Blackwellization estimator of the likelihood ratio.

  Potential
  Improvements:    More realistic penetrance function, improved sequential
                   imputation, handling unknown genetic models

  Acknowledgement: I am indebted to the many people who under Elizabeth
                   Thompson's mentorship, have produced a suite of library
                   modules known as MORGAN. I have reaped the benefits of many
                   years of development and experimentation. I would also like
                   to thank Myrna Jewett who has patiently answered my many
                   coding questions.

  Comments:        ability to handle binary trait with liability classes
                                                                    AWG 08/23/02

*******************************************************************************/

/**
 * @addtogroup lm_bayes
 * @{
 */

#define PNAM        "lm_bayes"
#define LM_BAYES

/* if there are liability classes */
/* MAKE SURE THE CORRECT penetrance function is set by allocate_loci()        */
/* set trait data discrete in parameter file -- e.g. bayes.par in Gold2       */

/* General included files.                                                    */
#include "lodscore.h"

/* Output file handling include files.                                        */
#include "./NewRtnes/output_handling.c"

/* M-H proposals.                                                             */
#include "./NewRtnes/mh_proposals.c"

/* MH restart.                                                                */
#include "./NewRtnes/mh_restart.c"

/* Pseudo-prior calculation.                                                  */
#include "./NewRtnes/norm_pseudo_prior.c"

int main (int argc, char *argv[])
{
   /* local variable declaration                                              */
   int   i, ii;                         /* indexes                            */
   int   cnt;                           /* counter                            */
   int   nsw;                           /* total number of swiches in
                                           pedigree file <= number of meioses */
   int   iteration;                     /* index iteration number of
                                           MCMC sampler                       */
   int   rao_freq = 0;                  /* count of iterations
                                           for the rao-Bl.. estimator         */
   double *pseudo_prior_calc;           /* used in calculating appropriate
                                           pseudo_priors                      */
   double *rao_realiz;                  /* Rao-Blackwellize estimate over
                                           all test positions for a specific
                                           realization                        */
   double **prob;                       /* required by the M sampler
                                           [num compnts][num loci]            */
   double p_linked;                     /* Prob of selecting the proposal
                                           which moves the trait locus along
                                           the linkage group as opposed to
                                           moving the trait locus to
                                           an unlinked position               */
   FILE *oSeedF;                        /* file pntr to seed file             */
   FILE *oRBF;                          /* file pntr to output file of
                                           rb estimates                       */

   /*  +-----------------------+                                              */
   /*  |   Begin program body  |                                              */
   /*  +=======================+                                              */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();
   prob = setupInherit (0);

   p_linked = .5;

   /* Allocate substructures associated with the global variable
      previous_realiz.                                                        */
   alloc_previous_realiz (PedSize, NMarks);

   /* Allocate pseudo-prior array.                                            */
   if (!InPseudo)
   {
      PseudoPrior = calloc (NTPos1, sizeof (double)); /* Array begins at 0,   */
      alloc_chk (WHERE, PseudoPrior);   /* unlinked as well.                  */

      /* Set pseudo-prior to unnormalized uniform.                            */
      for (ii = 0; ii < NTPos1; ii++)
         PseudoPrior[ii] = 1.0;

      /* Allocate pseudo-prior-calc array.                                    */
      pseudo_prior_calc = calloc (NTPos1, sizeof (double)); /* Array begins   */
      alloc_chk (WHERE, pseudo_prior_calc); /* at 0, unlinked as well.        */
   }

   /* Allocate rao_realiz array.                                              */
   rao_realiz = calloc (NTPos1, sizeof (double)); /* Array begins at 0,
                                                     unlinked as well.        */
   alloc_chk (WHERE, rao_realiz);

   /* Allocate rao blackwellization estimator array.                          */
   raoI_est = calloc (NTPos1, sizeof (double)); /* Array begins at 0.         */

   /* Allocate probability distributions for markers.                         */
   mrk_distr = malloc ((NMarks + 1) * sizeof (double)); /* Array begins at 1. */
   alloc_chk (WHERE, mrk_distr);

   /* Allocate array for frequency MCMC sampler visits each test position.    */
   tposit_freq = calloc (NTPos1, sizeof (int)); /* Array begins at 0.         */
   alloc_chk (WHERE, tposit_freq);

   /*   +-----------------------------------------------+                     */
   /*   |   ASSIGNING STARTING VALUES TO PARAMETERS     |                     */
   /*   +===============================================+                     */

   /*   Idea:  Using sequential imputation,                                   */
   /*          generate S_j^* and associated weights w_j                      */
   /*          (i.e. just single locus peeling).                              */
   /*          Then S^* and w^* are easily obtained.                          */
   /*          Choose an S^* as a starting value                              */
   /*          with prob \frac{w^*}{W} where W = sum w^*                      */

   /* Note - unlinked is interval 0, test position 0.                         */
   set_previous_realiz (0, 0, 0.5, 0.5, 0, xx0, xx1, NMarks, PedSize);

   prntStatusBMC ();                    /* Print status for beginning MCMC.   */

   /*   +---------------------------------------+                             */
   /*   |  BURN-IN                              |                             */
   /*   +=======================================+                             */

   /* Perform NBrnIt iterations (i.e. burn-in iterations) to allow the        */
   /* S's to mix correctly.  The trait locus remains unlinked.                */

   printf (" Beginning burn-in iterations \n");
   cnt = 1;                             /* Initialize counter.                */
   for (iteration = 1; iteration <= NBrnIt; iteration++)
   {
      if (!(iteration % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", iteration);
         fflush (stdout);
      }
      update_loci (LSamplProb, NMarks, PedSize, nsw, prob, nFGL);
   }
   printf ("\n");

   /*   +---------------------------------------+                             */
   /*   |  PSUEDO-PRIOR ASSIGNMENT              |                             */
   /*   +=======================================+                             */

/* To allow the sampler to explore regions with low posterior probability     */
/* a pseudo-prior (according to E(log P(Y_T | S_M))^-1) is also created.      */

   /* If the user has not specified pseudo-priors, calculate them by          */
   /* running the MCMC method with a uniform pseudo-prior.                    */
   if (!InPseudo)
   {
      printf (" Beginning pseudo-prior calculations ........ \n");

      /* Initialize marker prob distribution to uniform.  Change if
         one marker locus is to be favored.                                   */
      for (ii = 1; ii <= NMarks; ii++)
         mrk_distr[ii] = 1.0 / NMarks;

      rao_freq = 0;

      for (iteration = 1; iteration <= NPseudoPrIt; iteration++)
      {
         if (!(iteration % ChkIt))
         {
            printf (" Starting preliminary iteration %d\n", iteration);
            fflush (stdout);
         }

         update_loci (LSamplProb, NMarks, PedSize, nsw, prob, nFGL);

         if (previous_realiz.intvl == 0 && previous_realiz.rec_lft[FEM] == 0.5)
            /* Trait locus is presently unlinked.                             */
            update_unlnkd_2_lnkd (NMarks, PedSize, prob, nFGL, nCompnt,
                                  p_linked);

         else                           /* Trait locus is presently linked.   */
            update_position (p_linked, WindWdth, NMarks, PedSize, prob,
                             nFGL, nCompnt);

         if (!(iteration % ScorIt))     /* Calculate Rao-Bl estimator.        */
         {
            rao_freq++;
            calc_raoI (NMarks, rao_freq, rao_realiz);

            /* Using rao blackwellize estimator to calculate pseudo-priors.   */
            for (ii = 0; ii < NTPos1; ii++) /* Incremental mean.  Saves us
                                               from having to store all the
                                               Lange estimates.               */

               /* -1.0 appears below because we are taking inverse
                  log P(Y_T | S_M).                                           */
               pseudo_prior_calc[ii] = (rao_freq - 1) * pseudo_prior_calc[ii]
                  / rao_freq + (rao_realiz[ii]) / rao_freq;
         }

         if (SeqImpIt && !(iteration % SeqImpIt))
            mh_restart (NMarks, PedSize, prob, nCompnt, nFGL);

      }                                 /* END of LOOP                        */
      printf ("\n");

      normalize_pseudo_prior (pseudo_prior_calc); /* Normalize.               */

      /* Reset pseudo priors from uniform to the values based upon the
         rao-blackwellization estimates.                                      */
      for (ii = 0; ii < NTPos1; ii++)
         PseudoPrior[ii] = pseudo_prior_calc[ii];
   }

   /* Open output R-B estimates file, if requested.                           */

   if (OutRB)
      if (!(oRBF = open_fil (ORBEST)))
         give_up (ERR_qit);

   /*  +------------------------+                                             */
   /*  |  BEGIN MCMC SAMPLER    |                                             */
   /*  +========================+                                             */

   printf (" Beginning main MCMC sampler \n");

#if 0                                   /* Pseudo-priors are printed later.   */
   printf (" \n\n ");
   printf ("  PSEUDO-PRIORS\n\n");
   for (ii = 0; ii < NTPos1; ii++)
      printf ("pseudo_prb[%2d] = %4.2e\n", ii, PseudoPrior[ii]);
   printf ("\n");
#endif

   /* Initialize marker prob distribution to uniform.  Change if              */
   /* one marker locus is to be favored.                                      */
   for (ii = 1; ii <= NMarks; ii++)
      mrk_distr[ii] = 1.0 / NMarks;

   /* Re-initialize counters.                                                 */
   rao_freq = 0;
   for (ii = 0; ii < NTPos1; ii++)
   {
      tposit_freq[ii] = 0;
      raoI_est[ii] = 0.0;
   }

   for (iteration = 1; iteration <= NIt; iteration++)
   {
      if (!(iteration % ChkIt))
      {
         printf (" Starting main iteration %d\n", iteration);
         fflush (stdout);
      }

/* +------------------------------------------------+                         */
/* | UPDATE: S's WITHOUT A CHANGE OF TRAIT POSITION |                         */
/* +------------------------------------------------+                         */

      update_loci (LSamplProb, NMarks, PedSize, nsw, prob, nFGL);

      if (previous_realiz.intvl == 0 && previous_realiz.rec_lft[FEM] == 0.5)
         /* Trait locus is presently unlinked.                                */

/*   +--------------------------------------------------------------+         */
/*   | UPDATE: PROPOSE MOVING UNLINKED TRAIT LOCUS TO LINKED STATE  |         */
/*   +--------------------------------------------------------------+         */

         update_unlnkd_2_lnkd (NMarks, PedSize, prob, nFGL, nCompnt, p_linked);

      else                              /* Trait locus is presently linked.   */
         /*   +-------------------------------------------------------------+ */
         /*   |  UPDATE:  PROPOSE MOVING TRAIT LOCUS ALONG THE CHROMOSOME   | */
         /*   |           1. Move may be within marker interval.            | */
         /*   |           2. Move may be between marker interval.           | */
         /*   |           3. Move may be to UNLINKED STATE.                 | */
         /*   +-------------------------------------------------------------+ */
         update_position (p_linked, WindWdth, NMarks, PedSize, prob, nFGL,
                          nCompnt);

      if ((iteration % ScorIt) == 0)    /* Calculate Lange estimator.         */
      {
         rao_freq++;
         calc_raoI (NMarks, rao_freq, rao_realiz);
         if (OutRB)
            write_rb (oRBF, &rao_realiz);
      }

      if (SeqImpIt && !(iteration % SeqImpIt))
         mh_restart (NMarks, PedSize, prob, nCompnt, nFGL);

   }                                    /* END MCMC SAMPLER FOR LOOP */
   printf ("\n");

   prntStatusFMC ();                    /* Print status for MCMC completed.   */
   prntLodSHead ("LodScore estimates");

   prntLodSPos0 ();                     /* Unlinked trait position.          */
   printf ("%11.6f", PseudoPrior[0]);
   printf ("%9d", tposit_freq[0]);
   printf ("%10s%10s\n", "NA", "NA");

   for (i = 0, ii = 1; i < NTPos + NMarks; ++i)
   {
      if (!prntLodSPos (i))
         printf ("%11s%9s%10s%10s", "NA", "NA", "NA", "NA"); /* Marker pos.   */
      else
      {
         printf ("%11.6f", PseudoPrior[ii]);
         printf ("%9d", tposit_freq[ii]);
         if (tposit_freq[ii] == 0)      /* Test for test positions which are  */
            printf ("%10s", "NA");      /* not being visited.                 */
         else if (tposit_freq[0] == 0)  /* Unlinked case not visited.         */
            printf ("%10s", "INF");
         else
            printf ("%10.4f", log10 ((tposit_freq[ii] * PseudoPrior[0])
                                     / (tposit_freq[0] * PseudoPrior[ii])));
         printf ("%10.4f", log10 (raoI_est[ii] * PseudoPrior[0]
                                  / (raoI_est[0] * PseudoPrior[ii])));
         ++ii;
      }
      printf ("\n");
   }
   fflush (stdout);

/* Close output files.                                                        */

   if (OutRB)
      fclose (oRBF);

/* Free up memory.                                                            */
   free_xxgg ();
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (NMarks + 1, &prob);
   free_local (PedSize, NMarks);
   free_msmpl ();

/* Write new seed to seed file.                                               */
   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   exit (EXIT_SUCCESS);
}

/*  +-------------------------+                                               */
/*  |      SUBROUTINES        |                                               */
/*  +=========================+                                               */

void assign_xx (int nummrks, int nped)
/* Assign xx0 and xx1 in preparation for use in the L and M samplers.         */
{
   int   indvl, locus;

   for (locus = 0; locus <= nummrks; locus++)
      for (indvl = 0; indvl < nped; indvl++)
      {
         xx0[locus][indvl] = previous_realiz.xx0[locus][indvl];
         xx1[locus][indvl] = previous_realiz.xx1[locus][indvl];
      }

   return;
}

void alloc_previous_realiz (int nped, int nummrks)
{
   int   ii;

   previous_realiz.xx0 = malloc ((nummrks + 1) * sizeof (int *));
   alloc_chk (WHERE, previous_realiz.xx0);
   previous_realiz.xx1 = malloc ((nummrks + 1) * sizeof (int *));
   alloc_chk (WHERE, previous_realiz.xx1);
   previous_realiz.xx0[0] = malloc ((nped + 1) * (nummrks + 1) * sizeof (int));
   alloc_chk (WHERE, previous_realiz.xx0[0]);
   for (ii = 1; ii <= nummrks; ++ii)
      previous_realiz.xx0[ii] = previous_realiz.xx0[ii - 1] + (nped + 1);

   previous_realiz.xx1[0] = malloc ((nped + 1) * (nummrks + 1) * sizeof (int));
   alloc_chk (WHERE, previous_realiz.xx1[0]);
   for (ii = 1; ii <= nummrks; ++ii)
      previous_realiz.xx1[ii] = previous_realiz.xx1[ii - 1] + (nped + 1);

   return;
}

void free_local (int pedsize, int nummrks)
{
   free ((void *) PseudoPrior);
   free ((void *) mrk_distr);

   return;
}

/** @} */
