/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/**
 * This function builds global data structures used by the Autozyg programs.
 *
 * proc_auto() has as its input the results of parsing the parameter statements
 * for the Genedrop programs ibddrop and markerdrop, and for the Autozyg
 * programs, lm_auto, lm_ibdtests, lm_map and lm_pval.
 * proc_auto() uses many of the proc_drop() data structures and functions.
 * (Think of proc_auto as an extension of proc_drop.)
 * This function checks parameter statement data for consistency and complete-
 * ness and, if there are no errors, builds the global data structures.
 *
 * One chromosome only is selected.
 * At least 1 marker must be simulated (for ibddrop and markerdrop)
 * or selected (for lm_map, lm_ibdtests, lm_pval and lm_auto).
 * ibddrop simulates 1 tloc.
 *
 * Any combination of sex-not-specified maps or sex-specific maps may be used.
 *
 *      MAJ  revs 5/2000, 1/2002, 8/2003, 6/2005, 6/2006, 5/2007.
 */

int proc_auto (void)
{
   int   entry, sex, m_no, t_no, tl_no, loc_no;
   int   sxsp_maps = FALSE;             /* Whether or not sex-spec maps used. */
   int   tloc_maps = FALSE;             /* Whether or not there are tloc maps. */
   SIM_PTR p_s;
   PRNTMAP_PTR p_m;
   PRDIST_PTR p_l;

   if (!test_auto_sel ())               /* Checks simulate/select statement   */
      if (proc_err ())                  /* (fnc uses program names).          */
         return FALSE;

   if (n_err)
      return proc_end ();

   if (TRAIT_SELCTD && TRAIT_TLOC)      /* Finds the tloc for the selected    */
      if (!find_trait_tloc ())          /* trait.                             */
         if (proc_err ())
            return FALSE;

   if (INHER_SELCTD && INHER_TLOC)      /* Finds the tloc for the selected    */
      if (!find_inher_tloc ())          /* inheritance.                       */
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (TLOC_MAP)                        /* Set the tloc_maps flag.            */
      if ((TRAIT_SELCTD && TRAIT_TLOC) || TLOC_SELCTD
          || (INHER_SELCTD && INHER_TLOC))
         tloc_maps = TRUE;

   if (TRAIT_SELCTD && tloc_maps)       /* Finds the chromosome map (with     */
      if (!find_tr_tloc_map ())         /* unlinked specs) for the tloc for   */
         if (proc_err ())               /* the requested trait.               */
            return FALSE;

   if (TLOC_SELCTD && tloc_maps)        /* Finds the chromosome map (with     */
      if (!find_tloc_map ())            /* unlinked specs) for the requested  */
         if (proc_err ())               /* tloc.                              */
            return FALSE;

   if (INHER_SELCTD && tloc_maps)       /* Finds the chromosome map (with     */
      if (!find_inh_tloc_map ())        /* unlinked specs) for the tloc for   */
         if (proc_err ())               /* the requested inheritance.         */
            return FALSE;

   if (n_err)
      return proc_end ();

   if (tloc_maps)                       /* If there is a tloc map,            */
   {
      if (!updt_sim_lnkd_tloc ())       /* Adds the mapped tloc to chromosome */
         if (proc_err ())               /* to be simulated or selected.       */
            return FALSE;
      if (!updt_sim_unln_tloc ())       /* Adds the unlinked tloc to be       */
         if (proc_err ())               /* simulated or selected.             */
            return FALSE;
   }

   entry = FALSE;
   p_s = nxt_sim_chrom (&entry);        /* One chromosome only is selected.   */

   if (MARK_MAP)                        /* Checks sex-avrgd & sex-spcfc maps. */
   {
      if (!test_mark_map (p_s, &sxsp_maps)) /* Checks the marker map.         */
         if (proc_err ())
            return FALSE;
      if (tloc_maps)                    /* Checks the tloc map.               */
         if (!test_tloc_map (p_s, &sxsp_maps))
            if (proc_err ())
               return FALSE;
   }

   if (tloc_maps)                       /* For the simulated chromosome, if   */
   {                                    /* a mapped tloc is required, add to  */
      entry = FALSE;                    /* the tloc list.                     */
      while ((tl_no = nxt_sim_mapd_tl_no (&entry, p_s)))
         mak_tloc_list (tl_no, LNKD_TLOC);
      entry = FALSE;                    /* If an unlinked tloc is required,   */
      while ((tl_no = nxt_sim_unln_tl_no (&entry))) /* add to the simulated   */
         mak_tloc_list (tl_no, LNKD_TLOC); /* unlinked tloc list.             */
   }

   if (MARK_ALLELE)                     /* For each marker, check allele      */
   {                                    /* frequencies.                       */
      entry = FALSE;
      while ((m_no = nxt_sim_mark_no (&entry, p_s)))
         if (!test_mark_freq (p_s, m_no))
            if (proc_err ())
               return FALSE;
   }

   if (sim_traits && TLOC_ALLELE && !INHER_SELCTD) /* For each tloc, check    */
   {                                    /* allele frequencies.                */
      entry = FALSE;
      while ((tl_no = nxt_sim_tl_no (&entry)))
         if (!test_tloc_freq (tl_no))
            if (proc_err ())
               return FALSE;
   }

   entry = FALSE;
   t_no = nxt_sim_tr_no (&entry);       /* Get the trait number for the tloc. */

   if (sim_traits)
   {
      set_trait_typ (t_no);             /* Establish what type of trait is    */
      if (!test_trait_typ ())           /* selected.                          */
         if (proc_err ())
            return FALSE;
   }

   if (sim_traits && TRT_TLC_GENO)
   {
      entry = FALSE;
      while ((tl_no = nxt_tr_tl_no (&entry, t_no)))
         if (!test_tr_tl_genotyp (t_no, tl_no)) /* Check genotype means.      */
            if (proc_err ())
               return FALSE;
   }

   if (sim_traits && TRAIT_VARI)
      if (!test_trait_vars (t_no))      /* Check additive and residual        */
         if (proc_err ())               /* variances.                         */
            return FALSE;

   if (n_err)
      return proc_end ();

   if (MARK_MAP)                        /* Makes the "print map" for the sel  */
   {                                    /* markers and the tloc positions.
                                         * Assigns loci to sex-specific maps.
                                         * For maps not present, defaults to
                                         * AVG, then F or M maps.             */

      prnt_chrom_head (p_s);

      for (sex = FEM; sex <= AVG; ++sex) /* For each gender,                  */
      {
         p_m = mak_markprmap (p_s, sex); /* First, for the selected markers,  */

         if (MARK_MIN_DIST)
            if (!chk_markmindist (p_m)) /* Checks distances between markers,  */
               if (proc_err ())         /* if required.                       */
                  return FALSE;

         if (tloc_maps)                 /* Then, for sel tloc, if present.    */
            mak_tlocprmap (p_s, p_m, sex);

         shrink_prntmap (p_m);          /* Removes non selected marker loci.  */
         fix_prmap_dist (p_m);          /* Finds distances between adjacent   */
         /*                                markers for printing.              */

         if (tloc_maps)                 /* For each gender to be used,        */
            if ((sxsp_maps && (sex == FEM || sex == MAL))
                || (!sxsp_maps && sex == AVG))
               if (!chk_tlocprmap (p_m)) /* Checks tloc distances.            */
                  if (proc_err ())
                     return FALSE;
      }

      prnt_map (sxsp_maps);             /* Prints the chromosome map.         */
      if (PrintLevel >= 4)
         fprintf (stdout, "\n");

      if (MARK_ALLELE)                  /* Prints marker allele frequencies.  */
         prnt_mark_freqs (p_s);

      if (sim_traits && TLOC_ALLELE && !INHER_SELCTD)
         prnt_tloc_freqs (p_m);         /* Prints tloc allele frequencies.    */
   }

   if (tloc_maps)
      if (UNLTLOC_SELCTD)
      {
         prnt_unltloc_head ();          /* For the selected unmapped tloc,    */

         if (sim_traits && TLOC_ALLELE && !INHER_SELCTD)
            prnt_unltloc_freqs ();      /* prints the frequencies of unmapped */
         else
            prnt_unltloc_list ();       /* loci or names of unmapped loci.    */
      }

   if (INHER_SELCTD)                    /* Print the inheritance tloc list.   */
   {
      prnt_inher_head ();
      prnt_inh_tl_list ();
   }

   if (TRAIT_SELCTD)
   {                                    /* List traits and prints parameters. */
      prnt_trait_head ();

      if (TRT_TLC_GENO)
         prnt_tr_tl_genotyp ();
      else if (TRAIT_TLOC)
         prnt_tr_tl_list ();
      else if (!TRAIT_VARI)
         prnt_tr_list ();

      if (TRAIT_VARI)
         prnt_trait_vari ();

      if (PrintLevel >= 2)
         fprintf (stdout, "\n");
   }

   NMarks = p_m->no_mark_sel;           /* Counts loci.                       */
   NLnkdTL = p_m->no_tloc_sel;
   if (UNLTLOC_SELCTD)
      NUnlnTL = sim_unlnkd_tlocs->no_tloc_sel;
   else
      NUnlnTL = 0;

   NTLocs = NLnkdTL + NUnlnTL;
   NLoci = NMarks + 1;                  /* Now we count the tloc, whether or  */
   /*                                      not one is used.                   */

   if (!sim_traits)
      NTraits = 0;
   else
      NTraits = sim_traits->no_traits;

   if (tloc_maps)
   {
      entry = FALSE;
      while ((tl_no = nxt_sim_tl_no (&entry)))
         mak_tlocnam (tl_no);           /* Add this tloc name.                */
   }

   alloc_marek ();                      /* Stores marker recombination        */
   mak_marek ();                        /* fractions in Marek[][].            */
   mak_tarek ();                        /* Stores tloc recomb fract and pos   */
   /*                                      (or .5, .5 if no linked tloc).     */

   entry = FALSE;                       /* For each locus, stores allele      */
   p_m = nxt_prntmap (&entry, FEM);     /* frequencies in Qq[][].             */

   if (MARK_ALLELE || TLOC_ALLELE)      /* First markers and linked tloc      */
   {
      alloc_marsel ();
      loc_no = 0;

      entry = FALSE;
      while ((p_l = nxt_prntmap_locus (&entry, p_m)))
         if (p_l->locus_typ == TLOC_LOC && INHER_SELCTD)
            mak_allefrq (NULL, NULL, 0);
         else
         {
            mak_allefrq (p_m, p_l, p_l->no);
            if (p_l->locus_typ == MARK_LOC)
            {
               MarSel[loc_no] = p_l->no;
               ++loc_no;
            }
         }

      if (UNLTLOC_SELCTD)               /* or unlinked tloc,                  */
      {
         entry = FALSE;
         while ((tl_no = nxt_sim_unln_tl_no (&entry)))
            mak_allefrq (NULL, NULL, tl_no);
      }
      else if (!NTLocs)                 /* or if there is no tloc.            */
         mak_allefrq (NULL, NULL, 0);
   }

   if (MARK_NAME)
   {
      if (!test_marknam (p_s))          /* Checks marker names.               */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         if (mak_marknam (p_s))         /* Stores and prints marker names,    */
            prnt_marknam (p_s);         /* if provided.                       */
      }
      else
         return proc_end ();
   }

   if (MARK_DATA)
   {
      if (!test_markdat (p_s))          /* Checks marker data.                */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         mak_markdat (p_s);             /* Stores and prints marker data.     */
         prnt_markdat (p_s);
      }
      else
         return proc_end ();
   }

   if (PROBAND_GAM)
   {                                    /* Checks proband gametes and locus   */
      if (!test_probgam (p_s))          /* window.                            */
         if (proc_err ())
            return FALSE;

      if (PROB_GAM_WNDW)                /* Checks locus window for scoring    */
         if (!test_probgam_jnt (p_s))   /* pair of gametes.                   */
            if (proc_err ())
               return FALSE;

      if (PROB_GAM_PTRN)                /* Checks locus window for scoring    */
         if (!test_probgam_pat (p_s))   /* pattern labels.                    */
            if (proc_err ())
               return FALSE;

      if (!n_err)
      {
         mak_probgam ();                /* Stores proband gametes.            */
         if (NScCmp > 0)
         {
            mak_patlab ();              /* Stores window pattern labels.      */
            prnt_probgam ();            /* Prints locus window, proband       */
         }                              /* gametes and window patterns.       */
      }
   }

   if (!test_auto_pars ())              /* Check MC computational parameters. */
      if (proc_err ())
         return FALSE;

/* Stores incomplete penetrances, genotype means and variances.               */
   if (NTraits && (TraitTyp == QUANTITATIVE || TraitTyp == DISCRETE))
   {
      alloc_trait_pars (NTraits);
      entry = FALSE;
      t_no = nxt_sim_tr_no (&entry);    /* Get the trait number.              */
      pass_trait (t_no, 0);
   }

   if (TRAIT_DATA)                      /* Checks input record parameters.    */
      if (!test_ped_trtdat ())
         if (proc_err ())
            return FALSE;
   if (n_err)
      return proc_end ();

   mak_ped_dat ();

   if (TRAIT_DATA)                      /* Sets up and reports trait pedigree */
   {                                    /* data.                              */
      prnt_trt_typ ();
      prnt_ped_dat ();
   }

   if (err_typ)                         /* Handle any outstanding errors.     */
      if (proc_err ())
         return FALSE;

   if (TRAIT_PENE)                      /* Checks incomplete penetrances for  */
   {                                    /* discete trait data.                */
      if (!test_penetrnc ())
         if (proc_err ())
            return FALSE;

      prnt_penetrnc ();                 /* Prints penetrances.                */
   }

   if (TRAIT_DATA)                      /* Prints a blank line following the  */
      fprintf (stdout, "\n");           /* reporting of trait pedigree data.  */

   if (n_err)
      return proc_end ();

   prnt_auto_pars ();                   /* Prints MC parameters.              */

   return TRUE;
}
