/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 NAME
    log_prob_alt, log_prob_null

 SYNOPSIS

    #include "twoqtl.h"
    void log_prob_alt  (double *theta, int new, char ***S_trait_p,
                        char ***S_trait_m, int comp_no, int nsw,
                        int start, mc_eval_t *mc_eval_p);
    void log_prob_null (double *theta, int new, int comp_no, int nsw,
                        int start, mc_eval_t *mc_eval_p);

 DESCRIPTION
    Compute log Prob (traits | S_traits) for one component with
    realized S_traits and add it to mc_eval.

 INPUT

    S_trait_p   Paternal segregation indicators at trait positions.
    S_trait_m   Maternal segregation indicators at trait positions.
    comp_no     0, 1, ..., n_compnt - 1
    start       1 if current batch just starts, 0 otherwise
    nsw         Number of meioses that need to be updated.
    mc_eval

    theta, new are reserved for future use, not used right now.

 AUTHOR
     Written by Yun Ju Sung.

 HISTORY
     Documented.                                 Apr 15, 2005
     Extend to componentwise computation.        May 19, 2005
     Change to realize SI's at the null locus.   Jun 18, 2005
     Document revised.                           Jul  2, 2005

 SEE ALSO
     update_S_trait.c for allocating, realizing, and freeing S_trait,
     batch_means.c, twoqtl.h  for mc_eval.

*/

#include "twoqtl.h"

/* ALGORITHM for log_prob_alt:                                                */
/*   1. Pack S (S_p, S_m) for storing only changes in S                       */
/*      (not so much to save memory but to reduce computation later).         */
/*   2. Obtain the FGL of observed individuals at each locus in packed S.     */
/*   3. Compute Prob (Trait | FGL) with FGL at loci in packed S.              */
/*   4. Unpack S and update mc_eval.                                          */

void log_prob_alt (double *theta, int new, char ***S_trait_p,
                   char ***S_trait_m, int comp_no, int nsw,
                   int start, mc_eval_t * mc_eval_p)
{
   int  *map, map_size;
   int **FGL_p, **FGL_m;
   int **FGL_all;
   int  *FGL_a[2];                      /* Observed FGLs at the first locus.  */
   int  *FGL_b[2];                      /* Observed FGLs at the second locus. */
   int   packed_i;
   int   i1, i2;

   double **log_prob, *log_pr;
   double *from, *to_foo, *to_max;

   char **S_p = S_trait_p[comp_no];
   char **S_m = S_trait_m[comp_no];
   char **packed_S_p, **packed_S_m;

   /* Pack S.                                                                 */
   map = setup_map (S_p, S_m, comp_no, &map_size);
   packed_S_p = pack_S (S_p, map, map_size, comp_no);
   packed_S_m = pack_S (S_m, map, map_size, comp_no);

   if (DebugLevel > 1)
      print_step_1 (map, map_size, S_p, S_m, packed_S_p, packed_S_m, comp_no);

   /* Obtain FGL of observed individuals at each locus in packed S.           */

   FGL_p = (int **) malloc (map_size * sizeof (int *));
   alloc_chk (WHERE, FGL_p);

   FGL_m = (int **) malloc (map_size * sizeof (int *));
   alloc_chk (WHERE, FGL_m);

   for (i1 = 0; i1 < map_size; ++i1)
   {
      FGL_all = set_FGL_c (packed_S_p[i1], packed_S_m[i1], comp_no);
      FGL_p[i1] = obs_FGL (FGL_all[1], comp_no);
      FGL_m[i1] = obs_FGL (FGL_all[0], comp_no);
      free_2d ((void *) &FGL_all);
   }

   if (DebugLevel > 3)
      print_step_2 (FGL_p, FGL_m, map_size, comp_no);

   /* Compute log probability of trait given each observed FGL.               */

   log_prob = alloc_dbl2d (map_size, (NQtl == 1) ? 1 : map_size);

   for (i1 = 0; i1 < map_size; ++i1)
   {
      FGL_a[1] = FGL_p[i1];
      FGL_a[0] = FGL_m[i1];

      if (DebugLevel == 3)
         fprintf (stderr, "********* %2d\n", i1);

      if (NQtl == 1)
         log_prob[i1][0] = evaluate (FGL_a, NULL, comp_no);
      else
      {
         log_pr = log_prob[i1];
         for (i2 = 0; i2 < map_size; ++i2)
         {
            FGL_b[1] = FGL_p[i2];
            FGL_b[0] = FGL_m[i2];
            log_pr[i2] = evaluate (FGL_a, FGL_b, comp_no);
         }
      }
   }

   if (DebugLevel > 1)
      print_step_3 (log_prob, map_size);

   /* Update mc_eval, while unpacking S.                                      */

   n_loci_1 = mc_eval_p->n_loci_1;      /* n_loci in the original map.        */
   n_loci_2 = mc_eval_p->n_loci_2;

   for (i1 = 0; i1 < n_loci_1; ++i1)
   {
      packed_i = map[i1];               /* Locus in the packed map.           */
      from = log_prob[packed_i];
      to_foo = mc_eval_p->foo_alt[i1];
      to_max = mc_eval_p->max_alt[i1];

      for (i2 = 0; i2 < n_loci_2; ++i2)
         update_batch (from[map[i2]], start, to_foo + i2, to_max + i2);
   }

   free_2d ((void *) &log_prob);

   for (i1 = 0; i1 < map_size; ++i1)
   {
      free (FGL_p[i1]);
      free (FGL_m[i1]);
   }
   free (FGL_p);
   free (FGL_m);

   free (map);
   free_2d ((void *) &packed_S_p);
   free_2d ((void *) &packed_S_m);

   return;
}

void log_prob_null (double *theta, int new, int comp_no, int nsw, int start,
                    mc_eval_t * mc_eval_p)
{
   int **FGL_p, **FGL_m;
   int **FGL_all;
   int  *FGL_a[2];                      /* Observed FGLs at the first locus.  */
   int  *FGL_b[2];                      /* Observed FGLs at the second locus. */
   int   i1;

   double log_prob;
   char **S;

   FGL_p = (int **) malloc (NQtl * sizeof (int *));
   alloc_chk (WHERE, FGL_p);

   FGL_m = (int **) malloc (NQtl * sizeof (int *));
   alloc_chk (WHERE, FGL_m);

   for (i1 = 0; i1 < NQtl; ++i1)
   {
      S = update_S_null (comp_no);
      FGL_all = set_FGL_c (S[1], S[0], comp_no);
      FGL_p[i1] = obs_FGL (FGL_all[1], comp_no);
      FGL_m[i1] = obs_FGL (FGL_all[0], comp_no);

      if (DebugLevel > 2)
         print_step_2_null (S, FGL_all, FGL_p[i1], FGL_m[i1], i1, comp_no);

      free_2d ((void *) &FGL_all);
      free_2d ((void *) &S);
   }

   FGL_a[1] = FGL_p[0];
   FGL_a[0] = FGL_m[0];

   if (NQtl == 1)
      log_prob = evaluate (FGL_a, NULL, comp_no);
   else
   {
      FGL_b[1] = FGL_p[1];
      FGL_b[0] = FGL_m[1];
      log_prob = evaluate (FGL_a, FGL_b, comp_no);
   }

   if (DebugLevel > 2)
      fprintf (stderr, "null %e\n", log_prob);

   update_batch (log_prob, start, &(mc_eval_p->foo_null),
                 &(mc_eval_p->max_null));

   for (i1 = 0; i1 < NQtl; ++i1)
   {
      free (FGL_p[i1]);
      free (FGL_m[i1]);
   }
   free (FGL_p);
   free (FGL_m);

   return;
}

/* Map i1-th locus in S to locus in packed S.                                 */
int  *setup_map (char **S_p, char **S_m, int comp_no, int *size_p)
{
   char **S;

   int  *map;
   int   m_start, m_stop;
   int   n_loci, change, sex;
   int   m1, i1, i2;

   map = (int *) malloc (NLocPos * sizeof (int));
   alloc_chk (WHERE, map);

   m_start = mbase[comp_no];
   m_stop = m_start + n_meiosis[comp_no];

   n_loci = 0;
   map[0] = n_loci;                     /* Include the first locus.           */

   for (i1 = 1; i1 < NLocPos; ++i1)
   {
      change = 0;

      for (m1 = m_start; m1 < m_stop; ++m1)
      {
         sex = meiosis[m1].sex;
         i2 = comp_index ((meiosis[m1].ind).index, comp_no);

         S = (sex == 1) ? S_p : S_m;
         if (S[i1][i2] != S[i1 - 1][i2])
         {
            change = 1;
            break;
         }
      }

      if (change)
         map[i1] = ++n_loci;
      else
         map[i1] = n_loci;
   }

   *size_p = ++n_loci;                  /* n_loci in packed S.                */

   return map;
}

char **pack_S (char **S, int *map, int map_size, int comp_no)
{
   char **packed_S, *S_to, *S_from;

   int   comp_size;
   int   packed_i;
   int   i1, i2;

   comp_size = size_c[comp_no];

   packed_S = alloc_char2d (map_size, comp_size);

   packed_i = 0;

   for (i1 = 0; i1 < NLocPos; ++i1)
      if (i1 == 0 || (map[i1] != map[i1 - 1]))
      {
         S_to = packed_S[packed_i];
         S_from = S[i1];

         for (i2 = 0; i2 < comp_size; ++i2)
            S_to[i2] = S_from[i2];

         ++packed_i;
      }

   return packed_S;
}

/* Updated from set_FGL () in Sample.                                         */
int **set_FGL_c (char *S_p, char *S_m, int comp_no)
{
   int **FGL;
   int   comp_size;
   int   nghd_subscript;
   int   ma_index;
   int   pa_index;
   int   cg, i1;

   Ind  *dad, *mum;

   comp_size = size_c[comp_no];
   FGL = alloc_int2d (2, comp_size);
   cg = 0;

   for (i1 = 0; i1 < comp_size; ++i1)
   {
      nghd_subscript = nghd_index (i1, comp_no);
      dad = nghd[nghd_subscript].pa;

      if (S_m[i1] == -1 && isNullNode (dad))
      {
         ++cg;
         FGL[0][i1] = cg;
         ++cg;
         FGL[1][i1] = cg;
      }
      else
      {
         mum = nghd[nghd_subscript].ma;
         ma_index = comp_index (mum->index, comp_no);

         if ((S_m[i1] == -1) || (S_m[i1] == 0))
            FGL[0][i1] = FGL[0][ma_index];
         else
            FGL[0][i1] = FGL[1][ma_index];

         pa_index = comp_index (dad->index, comp_no);

         if ((S_p[i1] == -1) || (S_p[i1] == 0))
            FGL[1][i1] = FGL[0][pa_index];
         else
            FGL[1][i1] = FGL[1][pa_index];
      }
   }

   return FGL;
}

int  *obs_FGL (int *FGL, int comp_no)
{
   int  *obs_FGL;
   int   obs_size;
   int   i1;

   obs_size = obs_size_c[comp_no];

   obs_FGL = (int *) malloc (obs_size * sizeof (int));
   alloc_chk (WHERE, obs_FGL);

   for (i1 = 0; i1 < obs_size; ++i1)
      obs_FGL[i1] = FGL[obs_index[comp_no][i1]];

   return obs_FGL;
}

char **update_S_null (int comp_no)
{
   char **S_null, *S_p, *S_m, *S;

   int   comp_size;
   int   m_start, m_stop;
   int   sex;
   int   m1, i2;

   comp_size = size_c[comp_no];

   S_null = alloc_char2d (2, comp_size);
   S_p = S_null[1];
   S_m = S_null[0];

   memset (S_p, -1, comp_size * sizeof (char));
   memset (S_m, -1, comp_size * sizeof (char));

   m_start = mbase[comp_no];
   m_stop = m_start + n_meiosis[comp_no];

   for (m1 = m_start; m1 < m_stop; ++m1)
   {
      sex = meiosis[m1].sex;
      i2 = comp_index ((meiosis[m1].ind).index, comp_no);
      S = (sex == 0) ? S_m : S_p;
      S[i2] = (rans_alt () < .5) ? 0 : 1;
   }

   return S_null;
}

/* The following subroutines are printing routines called for various values  */
/* of DebugLevel at convenient locations in the preceding subroutines in this */
/* file to print different pieces of debugging and trace information.         */

void print_step_1 (int *map, int map_size, char **S_p, char **S_m,
                   char **packed_S_p, char **packed_S_m, int comp_no)
{
   int   i1;

   fprintf (stderr, "\n... Check step 1: component %2d ...\n", comp_no);

   if (DebugLevel > 2)
   {
      for (i1 = 0; i1 < NLocPos; ++i1)
         fprintf (stderr, "%2d %2d\n", i1, map[i1]);
   }

   if (map[NLocPos - 1] != map_size - 1)
      give_up_where (WHERE, ERR_qit);

   print_S (packed_S_p, packed_S_m, map_size, comp_no);

   if (DebugLevel == 3)
      print_S (S_p, S_m, NLocPos, comp_no);

   return;
}

void print_step_2 (int **FGL_p, int **FGL_m, int n_loci, int comp_no)
{
   int   locus;
   int   i1;

   fprintf (stderr, "\n... Check step 2: obs FGL for component %2d ...\n",
            comp_no);

   for (locus = 0; locus < n_loci; ++locus)
   {
      fprintf (stderr, "locus %3d", locus);
      for (i1 = 0; i1 < obs_size_c[comp_no]; ++i1)
         fprintf (stderr, " %3d%3d", FGL_p[locus][i1], FGL_m[locus][i1]);
      fprintf (stderr, "\n");
   }

   return;
}

/* The T_FGL_p and T_FGL_m arguments to the print_step_2_null subroutine are  */
/* single rows of the FGL_p and FGL_m variables declared in the log_prob_null */
/* subroutine.  They are specific to the locus being printed in the current   */
/* call to this subroutine.                                                   */

void print_step_2_null (char **S_n, int **FGL_all, int *T_FGL_p, int *T_FGL_m,
                        int locus_no, int comp_no)
{
   int   nghd_subscript;
   int   i1;

   fprintf (stderr, "\n... Check step 2: null locus %d component %2d ...\n",
            locus_no, comp_no);

   if (DebugLevel == 3)
   {
      for (i1 = 0; i1 < size_c[comp_no]; ++i1)
      {
         fprintf (stderr, "%2d %2d  %2d %2d ", (int) S_n[1][i1],
                  (int) S_n[0][i1], FGL_all[1][i1], FGL_all[0][i1]);
         if (locus_no < NTLocs)
         {
            nghd_subscript = nghd_index (i1, comp_no);
            fprintf (stderr, "%2d %2d\n", gg1[0][nghd_subscript],
                     gg0[0][nghd_subscript]);
         }
         else
            fprintf (stderr, "\n");
      }
   }

   for (i1 = 0; i1 < obs_size_c[comp_no]; ++i1)
      fprintf (stderr, "%2d %2d  %2d %2d\n", i1, obs_index[comp_no][i1],
               T_FGL_p[i1], T_FGL_m[i1]);

   return;
}

void print_step_3 (double **log_prob, int map_size)
{
   double *log_pr;

   int   stop;
   int   i1, i2;

   fprintf (stderr, "\n... Check step 3: log probability ...\n");

   stop = (NQtl == 1) ? 1 : map_size;

   for (i1 = 0; i1 < map_size; ++i1)
   {
      log_pr = log_prob[i1];

      fprintf (stderr, "%2d ", i1);
      for (i2 = 0; i2 < stop; ++i2)
         fprintf (stderr, "%e ", log_pr[i2]);
      fprintf (stderr, "\n");
   }

   return;
}
